!function(e,t){"object"==typeof exports&&"object"==typeof module?module.exports=t():"function"==typeof define&&define.amd?define([],t):"object"==typeof exports?exports.Taipy=t():e.Taipy=t()}(self,(function(){return function(){"use strict";var e=function(e,t,n,o,a,r,s,p,i){var f=new XMLHttpRequest;f.open("POST","".concat(t,"?client_id=").concat(p),!1),f.onerror=function(e){return self.postMessage({message:"Error: "+e,error:!0})},f.onload=function(e){return i(e.lengthComputable?e.loaded:0)};var l=new FormData;l.append("blob",e,r),l.append("part",o.toString()),l.append("total",a.toString()),l.append("var_name",n),l.append("multiple",s?"True":"False"),f.send(l)},t=1048576,n=function(e,t){return function(n){return self.postMessage({progress:100*(t+n)/e,done:!1})}};return self.onmessage=function(o){!function(o,a,r,s){if(o){for(var p=0,i=0;i<o.length;i++)p+=o[i].size;var f=[];for(i=0;i<o.length;i++)for(var l=o[i],u=l.size,d=0,c=t,g=Math.ceil(u/t);d<u;){var m=l.slice(d,c),v=n(p,d);v(0),e(m,a,r,Math.floor(d/t),g,l.name,0!=i&&o.length>0,s,v),v(m.size),c=(d=c)+t,f.push(l.name)}self.postMessage({progress:100,message:f.join(", ")+" Uploaded Succesfully",done:!0})}}(o.data.files,o.data.uploadUrl,o.data.varName,o.data.id)},{}}()}));